/* $Id: x86_setup.c,v 1.5.2.1 2001/06/17 15:35:20 mikpe Exp $
 * Performance-monitoring counters driver.
 * x86-specific kernel-resident code.
 *
 * Copyright (C) 1999-2001  Mikael Pettersson
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <asm/processor.h>
#include <asm/perfctr.h>
#include "compat.h"
#include "x86_compat.h"
#include <linux/perfctr.h>	/* for DEBUG */

void __init perfctr_dodgy_tsc(void)
{
#if defined(CONFIG_PERFCTR_WINCHIP)
	/* Check for Centaur WinChip C6/2/3 with broken TSC */
	if( boot_cpu_data.x86_vendor != X86_VENDOR_CENTAUR )
		return;
	if( boot_cpu_data.x86 != 5 )
		return;
	switch( boot_cpu_data.x86_model ) {
	case 4:	break;		/* WinChip C6 */
	case 8:	break;		/* WinChip 2 */
	case 9: break;		/* WinChip 3 */
	default: return;
	}
	/*
	 * WinChip processors implement the TSC by using the
	 * low 32 bits of the two performance counters.
	 * If the performance counters are reprogrammed, the
	 * RDTSC instruction will yield incorrect results.
	 * (On the C6, a read of the TSC using RDTSC or RDMSR
	 * may even cause the processor to hang in this case.
	 * See erratum I-13 in the C6 data sheet for details.)
	 *
	 * Since the user explicitly requested the performance
	 * counters, we have to disable all uses of the TSC.
	 */
	set_cr4_tsd_local();
	clr_cap_tsc();
	printk(KERN_NOTICE "WinChip Time-Stamp Counter disabled\n");
#endif
}

#if defined(CONFIG_PERFCTR_DEBUG) && defined(CONFIG_PERFCTR_VIRTUAL)
struct shadow_vperfctr {
	unsigned int pad[512];
	struct thread_struct *thread;
	struct shadow_vperfctr *next;
	unsigned int magic;
#define VPERFCTR_MAGIC 0x00010006	/* binary layout V1.6 */
};
static spinlock_t shadow_vp_list_lock = SPIN_LOCK_UNLOCKED;
static struct shadow_vperfctr *shadow_vp_list;

void _vperfctr_set_shadow(struct thread_struct *thread, struct vperfctr *perfctr)
{
	struct shadow_vperfctr *here, **pprev;

	spin_lock(&shadow_vp_list_lock);
	pprev = &shadow_vp_list;
	if( !perfctr ) {
		while( (here = *pprev) != NULL ) {
			if( here->thread == thread ) {
				*pprev = here->next;
				here->next = NULL;
				here->thread = NULL;
				break;
			}
			pprev = &here->next;
		}
	} else {
		here = (struct shadow_vperfctr*)perfctr;
		here->thread = thread;
		here->next = *pprev;
		here->magic = VPERFCTR_MAGIC;
		*pprev = here;
	}
	spin_unlock(&shadow_vp_list_lock);
}

static struct vperfctr *__vperfctr_get_shadow(const struct thread_struct *thread)
{
	struct shadow_vperfctr *vp;

	spin_lock(&shadow_vp_list_lock);
	for(vp = shadow_vp_list; vp && vp->thread != thread; vp = vp->next)
		;
	spin_unlock(&shadow_vp_list_lock);
	return (struct vperfctr*)vp;
}

struct vperfctr *_vperfctr_get_shadow(const struct thread_struct *thread,
				      const char *function)
{
	struct vperfctr *perfctr = thread->perfctr;
	struct vperfctr *shadow = __vperfctr_get_shadow(thread);
	if( perfctr != shadow ) {
		printk(KERN_ERR "%s: BUG! perfctr has been clobbered, is 0x%08lx should be 0x%08lx (pid %d, comm %s)\n",
		       function, (long)perfctr, (long)shadow, current->pid, current->comm);
		return NULL;
	}
	if( (long)perfctr & (4096-1) ) {
		printk(KERN_ERR "%s: BUG! perfctr 0x%08lx is not page aligned (pid %d, comm %s)\n",
		       function, (long)perfctr, current->pid, current->comm);
		return NULL;
	}
	if( perfctr && ((struct shadow_vperfctr*)perfctr)->magic != VPERFCTR_MAGIC ) {
		printk(KERN_ERR "%s: BUG! perfctr 0x%08lx has invalid magic 0x%08x\n (pid %d, comm %s)\n",
		       function, (long)perfctr, ((struct shadow_vperfctr*)perfctr)->magic, current->pid, current->comm);
		return NULL;
	}
	return perfctr;
}

#ifdef CONFIG_PERFCTR_MODULE
EXPORT_SYMBOL(_vperfctr_set_shadow);
EXPORT_SYMBOL(_vperfctr_get_shadow);
#endif

#endif	/* DEBUG && VIRTUAL */

#ifdef CONFIG_PERFCTR_MODULE
MAYBE_EXPORT_mmu_cr4_features;
EXPORT_cpu_khz;
#endif
